"""Módulo de Método de Diferencias Finitas (MDF)

Este submódulo contiene implementaciones para resolver EDPs utilizando el
método de diferencias finitas. Actualmente, incluye un solucionador directo
para la ecuación de Laplace en mallas rectangulares.

"""
from .mdf import DiferenciasFinitas